#include <cstdlib>
#include <utility>
#include <iostream>
#include <iomanip>
#include <gmpxx.h>

namespace My {
    const size_t width = 17;
    class Error_zero {};
    inline bool judge_sign_of(const mpz_class &u, const mpz_class &v, const mpz_class &qs);
    inline bool under_or_over(const mpz_class &uu, const mpz_class &threshhold, const mpz_class &v, const mpz_class &qs);
    inline bool and_or(const bool a, const bool b, const mpz_class &v, const mpz_class &qs);
    void print_the_result_of_an_iteration(const int j, const mpz_class &a, const mpz_class &u, const mpz_class &v);
}

// If v > 0 and qs > 0, then ask whether u <  0.
// If v > 0 and qs < 0, then ask whether u <= 0.
// If v < 0 and qs > 0, then ask whether u >= 0.
// If v < 0 and qs < 0, then ask whether u >  0.
inline bool My::judge_sign_of(const mpz_class &u, const mpz_class &v, const mpz_class &qs)
{
    if (v == 0 || qs == 0) throw Error_zero();
    return v >= 0
        ? (qs >= 0 ? u <  0 : u <= 0)
        : (qs >= 0 ? u >= 0 : u >  0);
}

// If v and qs share the same sign, then ask whether uu < threshhold.
// If v and qs have opposite signs, then ask whether uu > threshhold.
inline bool My::under_or_over(const mpz_class &uu, const mpz_class &threshhold, const mpz_class &v, const mpz_class &qs)
{
    if (v == 0 || qs == 0) throw Error_zero();
    return (v >= 0 && qs >= 0) || (v < 0 && qs < 0)
        ? uu < threshhold
        : uu > threshhold;
}

// If v and qs share the same sign, then return a || b.
// If v and qs have opposite signs, then return a && b.
inline bool My::and_or(const bool a, const bool b, const mpz_class &v, const mpz_class &qs)
{
    if (v == 0 || qs == 0) throw Error_zero();
    return (v >= 0 && qs >= 0) || (v < 0 && qs < 0)
        ? a || b
        : a && b;
}

void My::print_the_result_of_an_iteration(const int j, const mpz_class &a, const mpz_class &u, const mpz_class &v)
{
    if (j < 0) {
        std::cout
            << std::setw(width) << "" << " "
            << std::setw(width) << "" << " "
            << std::setw(width) << u  << " "
            << std::setw(width) << v  << "\n";
    }
    else {
        std::cout
            << std::setw(width) << j  << " "
            << std::setw(width) << a  << " "
            << std::setw(width) << u  << " "
            << std::setw(width) << v  << "\n";
    }
}

int main(const int argc, const char *const *argv) {

    if (argc != 6) {
        std::cerr << "usage: " << argv[0] << " no_iter n s p q\n";
        std::exit(1);
    }
    const int no_iter {std::atoi(argv[1])};
    const mpz_class n {std::atoi(argv[2])};
    const mpz_class s {std::atoi(argv[3])};
    const mpz_class p {std::atoi(argv[4])};
    const mpz_class q {std::atoi(argv[5])};
    {
        bool are_args_impossible {false};
        if (no_iter < 0 || n < 0 || q == 0) are_args_impossible = true;
        if (!are_args_impossible && n != 0 && s != 0) {
            mpz_class candidate_root {0};
            mpz_class candidate      {0};
            while (candidate < n) {
                ++candidate_root;
                candidate = candidate_root * candidate_root;
            }
            if (candidate == n) are_args_impossible = true;
        }
        if (are_args_impossible) {
            std::cerr << argv[0] << " error: impossible arguments\n";
            std::exit(1);
        }
    }

    if (n == 0 || s == 0) {
        mpz_class u {p};
        mpz_class v {q};
        My::print_the_result_of_an_iteration(-1, 0, u, v);
        for (int j {0}; j < no_iter; ++j) {
            mpz_class a {0};
            while (true) {
                // If v > 0, then break if u < v.
                // If v < 0, then break if u > v.
                if (My::under_or_over(u, v, v, 1)) break;
                u -= v;
                ++a;
            }
            while (true) {
                const mpz_class u_new = u + v;
                // If v > 0, then break if u_new >= v.
                // If v < 0, then break if u_new <= v.
                if (!My::under_or_over(u_new, v, v, 1)) break;
                u = u_new;
                --a;
            }
            swap(u, v);
            My::print_the_result_of_an_iteration(j, a, u, v);
            if (v == 0) break;
        }
    }
    else {
        const mpz_class qs {q * s};
        mpz_class        u {q * p};
        mpz_class        v {q * q};
        std::cout
            << std::setw(My::width) << "" << " "
            << std::setw(My::width) << "" << " "
            << std::setw(My::width) << u  << " "
            << std::setw(My::width) << v  << "\n";
        for (int j {0}; j < no_iter; ++j) {
            const mpz_class threshhold {qs * qs * n};
            mpz_class                a {0};
            u = -u;
            while (true) {
                const mpz_class uu = u * u;
                // If v > 0 and qs > 0, then break if u <  0 || uu < threshhold.
                // If v > 0 and qs < 0, then break if u <= 0 && uu > threshhold.
                // If v < 0 and qs > 0, then break if u >= 0 && uu > threshhold.
                // If v < 0 and qs < 0, then break if u >  0 || uu < threshhold.
                if (My::and_or(
                    My::judge_sign_of(u, v, qs),
                    My::under_or_over(uu, threshhold, v, qs),
                    v, qs
                )) break;
                u -= v;
                --a;
            }
            while (true) {
                const mpz_class u_new = u + v;
                const mpz_class uu = u_new * u_new;
                // If v > 0 and qs > 0, then break if u_new >= 0 && uu >= threshhold.
                // If v > 0 and qs < 0, then break if u_new >  0 || uu <= threshhold.
                // If v < 0 and qs > 0, then break if u_new <  0 || uu <= threshhold.
                // If v < 0 and qs < 0, then break if u_new <= 0 && uu >= threshhold.
                if (My::and_or(
                    My::judge_sign_of(u_new, -v, qs),
                    !My::under_or_over(uu, threshhold, v, qs),
                    v, -qs
                )) break;
                u = u_new;
                ++a;
            }
            {
                const mpz_class numerator {-u*u + threshhold};
                mpz_class        quotient {0};
                mpz_class       remainder {0};
                mpz_tdiv_qr(quotient.get_mpz_t(), remainder.get_mpz_t(), numerator.get_mpz_t(), v.get_mpz_t());
                if (remainder != 0) throw My::Error_zero();
                v = quotient;
            }
            std::cout
                << std::setw(My::width) << j << " "
                << std::setw(My::width) << a << " "
                << std::setw(My::width) << u << " "
                << std::setw(My::width) << v << "\n";
        }
    }

    return 0;

}

